# stdlib
from typing import Optional

# 3rd party
from domdf_python_tools.paths import PathPlus
from sphinx.application import Sphinx
from sphinx_toolbox import latex


def replace_emoji(app: Sphinx, exception: Optional[Exception] = None):
	if exception:
		return

	if app.builder.name.lower() != "latex":
		return

	output_file = PathPlus(app.builder.outdir) / f"{app.builder.titles[0][1]}.tex"
	output_content = output_file.read_text()
	output_content = output_content.replace('😉', "\\Smiley")
	output_file.write_clean(output_content)


def setup(app: Sphinx):
	app.connect("build-finished", replace_emoji)
	app.connect("build-finished", latex.replace_unknown_unicode)



# stdlib
from typing import List

# 3rd party
import pytest
from coincidence.regressions import AdvancedDataRegressionFixture
from consolekit.terminal_colours import Fore

# this package
from shippinglabel.classifiers import classifiers_from_requirements, validate_classifiers
from shippinglabel.requirements import ComparableRequirement


class TestValidateClassifiers:

	def test_errors(self, capsys):
		validate_classifiers(["Foo :: Bar", "Foo :: Bar :: Baz", "Fuzzy :: Wuzzy :: Was :: A :: Bear"])
		captured = capsys.readouterr()

		stderr = captured.err.split('\n')
		assert stderr[0].endswith(f"Unknown Classifier 'Foo :: Bar'!{Fore.RESET}")
		assert stderr[1].endswith(f"Unknown Classifier 'Foo :: Bar :: Baz'!{Fore.RESET}")
		assert stderr[2].endswith(f"Unknown Classifier 'Fuzzy :: Wuzzy :: Was :: A :: Bear'!{Fore.RESET}")
		assert not captured.out

	def test_deprecated(self, capsys):
		validate_classifiers(["Natural Language :: Ukranian"])
		captured = capsys.readouterr()

		stderr = captured.err.split('\n')
		assert stderr[0].endswith(f"Classifier 'Natural Language :: Ukranian' is deprecated!{Fore.RESET}")
		assert not captured.out

	def test_valid(self, capsys):
		validate_classifiers(["Natural Language :: Ukrainian", "License :: OSI Approved"])
		captured = capsys.readouterr()
		assert not captured.out
		assert not captured.err


@pytest.mark.parametrize(
		"requirements",
		[
				pytest.param(["dash"], id="dash_upper"),
				pytest.param(["Dash"], id="dash_lower"),
				pytest.param(["jupyter"], id="jupyter_upper"),
				pytest.param(["Jupyter"], id="jupyter_lower"),
				pytest.param(["matplotlib"], id="matplotlib_upper"),
				pytest.param(["Matplotlib"], id="matplotlib_lower"),
				pytest.param(["pygame"], id="pygame"),
				pytest.param(["arcade"], id="arcade"),
				pytest.param(["flake8"], id="flake8"),
				pytest.param(["flake8-walrus"], id="flake8-walrus"),
				pytest.param(["flask"], id="flask"),
				pytest.param(["werkzeug"], id="werkzeug"),
				pytest.param(["click>=2.0,!=2.0.1"], id="click"),
				pytest.param(["pytest"], id="pytest"),
				pytest.param(["pytest-randomly"], id="pytest-randomly"),
				pytest.param(["tox"], id="tox"),
				pytest.param(["tox-envlist"], id="tox-envlist"),
				pytest.param(["Tox==1.2.3"], id="Tox==1.2.3"),
				pytest.param(["sphinx"], id="sphinx"),
				pytest.param(["sphinx-toolbox"], id="sphinx-toolbox"),
				pytest.param(["Sphinx>=1.3"], id="Sphinx>=1.3"),
				pytest.param(["jupyter", "matplotlib>=3"], id="jupyter_matplotlib"),
				pytest.param(["dash", "flask"], id="dash_flask"),
				pytest.param(["pytest", "flake8"], id="pytest_flake8"),
				pytest.param(["dulwich"], id="dulwich"),
				pytest.param(["gitpython"], id="gitpython"),
				pytest.param(["dulwich", "southwark"], id="dulwich_southwark"),
				]
		)
def test_classifiers_from_requirements(
		requirements: List[str],
		advanced_data_regression: AdvancedDataRegressionFixture,
		):
	requirements_list = [ComparableRequirement(req) for req in requirements]
	advanced_data_regression.check(list(classifiers_from_requirements(requirements_list)))



# stdlib
import os
from typing import List

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike

# this package
from shippinglabel import normalize
from shippinglabel.requirements import (
		ComparableRequirement,
		RequirementsManager,
		combine_requirements,
		read_requirements
		)


class DocRequirementsManager(RequirementsManager):
	target_requirements = {
			ComparableRequirement("sphinxcontrib-httpdomain>=1.7.0"),
			ComparableRequirement("sphinxemoji>=0.1.6"),
			ComparableRequirement("sphinx-notfound-page>=0.5"),
			ComparableRequirement("sphinx-tabs>=1.1.13"),
			ComparableRequirement("autodocsumm>=0.2.0"),
			# ComparableRequirement("sphinx-gitstamp"),
			# ComparableRequirement("gitpython"),
			# ComparableRequirement("sphinx_autodoc_typehints>=1.11.0"),
			ComparableRequirement("sphinx-copybutton>=0.2.12"),
			ComparableRequirement("sphinx-prompt>=1.1.0"),
			ComparableRequirement("sphinx>=3.0.3"),
			}

	def __init__(self, repo_path: PathLike):
		self.filename = os.path.join("doc-source", "requirements.txt")
		super().__init__(repo_path)

	def compile_target_requirements(self) -> None:
		# Mapping of pypi_name to version specifier
		theme_versions = {
				"alabaster": ">=0.7.12",
				"sphinx_rtd_theme": "<0.5",
				"domdf_sphinx_theme": ">=0.1.0",
				"repo_helper_sphinx_theme": ">=0.0.2",
				}

		for name, specifier in theme_versions.items():
			if name == "alabaster":
				self.target_requirements.add(ComparableRequirement(f"{name}{specifier}"))
				break
		else:
			self.target_requirements.add(ComparableRequirement("alabaster>=0.7.12"))

		# Mapping of pypi_name to version specifier
		my_sphinx_extensions = {
				"extras_require": ">=0.2.0",
				"seed_intersphinx_mapping": ">=0.1.1",
				"default_values": ">=0.2.0",
				"toctree_plus": ">=0.0.4",
				"sphinx-toolbox": ">=1.7.1",
				}

		for name, specifier in my_sphinx_extensions.items():
			if name != "shippinglabel":
				self.target_requirements.add(ComparableRequirement(f"{name}{specifier}"))

	def merge_requirements(self) -> List[str]:
		current_requirements, comments = read_requirements(self.req_file)

		for req in current_requirements:
			req.name = normalize(req.name)
			if req.name not in self.get_target_requirement_names():
				self.target_requirements.add(req)

		self.target_requirements = set(combine_requirements(self.target_requirements))

		return comments


def test_requirements_manager(tmp_pathplus: PathPlus):

	(tmp_pathplus / "requirements.txt").write_text('')
	(tmp_pathplus / "doc-source").mkdir()
	(tmp_pathplus / "doc-source" / "requirements.txt").write_text('')

	DocRequirementsManager(tmp_pathplus).run()

	assert (tmp_pathplus / "doc-source" / "requirements.txt").read_lines() == [
			"alabaster>=0.7.12",
			"autodocsumm>=0.2.0",
			"default-values>=0.2.0",
			"extras-require>=0.2.0",
			"seed-intersphinx-mapping>=0.1.1",
			"sphinx>=3.0.3",
			"sphinx-copybutton>=0.2.12",
			"sphinx-notfound-page>=0.5",
			"sphinx-prompt>=1.1.0",
			"sphinx-tabs>=1.1.13",
			"sphinx-toolbox>=1.7.1",
			"sphinxcontrib-httpdomain>=1.7.0",
			"sphinxemoji>=0.1.6",
			"toctree-plus>=0.0.4",
			'',
			]

	with (tmp_pathplus / "doc-source" / "requirements.txt").open('a', encoding="UTF-8") as fp:
		fp.write("lorem>=0.1.1")

	DocRequirementsManager(tmp_pathplus).run()

	assert (tmp_pathplus / "doc-source" / "requirements.txt").read_lines() == [
			"alabaster>=0.7.12",
			"autodocsumm>=0.2.0",
			"default-values>=0.2.0",
			"extras-require>=0.2.0",
			"lorem>=0.1.1",
			"seed-intersphinx-mapping>=0.1.1",
			"sphinx>=3.0.3",
			"sphinx-copybutton>=0.2.12",
			"sphinx-notfound-page>=0.5",
			"sphinx-prompt>=1.1.0",
			"sphinx-tabs>=1.1.13",
			"sphinx-toolbox>=1.7.1",
			"sphinxcontrib-httpdomain>=1.7.0",
			"sphinxemoji>=0.1.6",
			"toctree-plus>=0.0.4",
			'',
			]


def test_requirements_manager_extras(tmp_pathplus: PathPlus):

	(tmp_pathplus / "requirements.txt").write_text("domdf_python_tools>=1.5.0")
	(tmp_pathplus / "doc-source").mkdir()
	(tmp_pathplus / "doc-source" / "requirements.txt").write_text("domdf_python_tools[testing]>=1.5.0")

	DocRequirementsManager(tmp_pathplus).run()

	assert (tmp_pathplus / "doc-source" / "requirements.txt").read_lines() == [
			"alabaster>=0.7.12",
			"autodocsumm>=0.2.0",
			"default-values>=0.2.0",
			"domdf-python-tools[testing]>=1.5.0",
			"extras-require>=0.2.0",
			"seed-intersphinx-mapping>=0.1.1",
			"sphinx>=3.0.3",
			"sphinx-copybutton>=0.2.12",
			"sphinx-notfound-page>=0.5",
			"sphinx-prompt>=1.1.0",
			"sphinx-tabs>=1.1.13",
			"sphinx-toolbox>=1.7.1",
			"sphinxcontrib-httpdomain>=1.7.0",
			"sphinxemoji>=0.1.6",
			"toctree-plus>=0.0.4",
			'',
			]



#!/usr/bin/env python3

# This file is managed by 'repo_helper'. Don't edit it directly.

# stdlib
import os
import re
import sys

# 3rd party
from sphinx_pyproject import SphinxConfig

sys.path.append('.')

config = SphinxConfig(globalns=globals())
project = config["project"]
author = config["author"]
documentation_summary = config.description

github_url = "https://github.com/{github_username}/{github_repository}".format_map(config)

rst_prolog = f""".. |pkgname| replace:: shippinglabel
.. |pkgname2| replace:: ``shippinglabel``
.. |browse_github| replace:: `Browse the GitHub Repository <{github_url}>`__
"""

slug = re.sub(r'\W+', '-', project.lower())
release = version = config.version

sphinx_builder = os.environ.get("SPHINX_BUILDER", "html").lower()
todo_include_todos = int(os.environ.get("SHOW_TODOS", 0)) and sphinx_builder != "latex"

intersphinx_mapping = {
		"python": ("https://docs.python.org/3/", None),
		"sphinx": ("https://www.sphinx-doc.org/en/stable/", None),
		"requests": ("https://requests.readthedocs.io/en/latest/", None),
		"pypi-json": ("https://pypi-json.readthedocs.io/en/latest/", None),
		"apeye": ("https://apeye.readthedocs.io/en/latest/", None),
		}

html_theme_options = {"logo_only": False}

html_context = {
		"display_github": True,
		"github_user": "domdfcoding",
		"github_repo": "shippinglabel",
		"github_version": "master",
		"conf_py_path": "/doc-source/",
		}
htmlhelp_basename = slug

latex_documents = [("index", f'{slug}.tex', project, author, "manual")]
man_pages = [("index", slug, project, [author], 1)]
texinfo_documents = [("index", slug, project, author, slug, project, "Miscellaneous")]

toctree_plus_types = set(config["toctree_plus_types"])

autodoc_default_options = {
		"members": None,  # Include all members (methods).
		"special-members": None,
		"autosummary": None,
		"show-inheritance": None,
		"exclude-members": ','.join(config["autodoc_exclude_members"]),
		}

latex_elements = {
		"printindex": "\\begin{flushleft}\n\\printindex\n\\end{flushleft}",
		"tableofcontents": "\\pdfbookmark[0]{\\contentsname}{toc}\\sphinxtableofcontents",
		}


# Fix for pathlib issue with sphinxemoji on Python 3.9 and Sphinx 4.x
def copy_asset_files(app, exc):
	# 3rd party
	from domdf_python_tools.compat import importlib_resources
	from sphinx.util.fileutil import copy_asset

	if exc:
		return

	asset_files = ["twemoji.js", "twemoji.css"]
	for path in asset_files:
		path_str = os.fspath(importlib_resources.files("sphinxemoji") / path)
		copy_asset(path_str, os.path.join(app.outdir, "_static"))


def setup(app):
	# 3rd party
	from sphinx_toolbox.latex import better_header_layout
	from sphinxemoji import sphinxemoji

	app.connect("config-inited", lambda app, config: better_header_layout(config))
	app.connect("build-finished", copy_asset_files)
	app.add_js_file("https://unpkg.com/twemoji@latest/dist/twemoji.min.js")
	app.add_js_file("twemoji.js")
	app.add_css_file("twemoji.css")
	app.add_transform(sphinxemoji.EmojiSubstitutions)


html_logo = "../shippinglabel.png"
latex_logo = "../shippinglabel.png"
latex_elements["preamble"] = "\\usepackage{marvosym}"
nitpicky = True
needspace_amount = r"5\baselineskip"
ignore_missing_xrefs = ["^packaging.utils.InvalidSdistFilename$", "^wheel_filename"]



# stdlib
from typing import Callable, Type, TypeVar

# 3rd party
from apeye.url import URL
from betamax import Betamax  # type: ignore
from domdf_python_tools.paths import PathPlus
from packaging.tags import Tag
from pytest_regressions.data_regression import RegressionYamlDumper

pytest_plugins = ("coincidence", )

_C = TypeVar("_C", bound=Callable)

with Betamax.configure() as config:
	config.cassette_library_dir = PathPlus(__file__).parent / "cassettes"


def _representer_for(*data_type: Type):  # noqa: MAN002

	def deco(representer_fn: _C) -> _C:
		for dtype in data_type:
			RegressionYamlDumper.add_custom_yaml_representer(dtype, representer_fn)

		return representer_fn

	return deco


@_representer_for(URL, Tag)
def _represent_sequences(dumper: RegressionYamlDumper, data):  # noqa: MAN001,MAN002
	return dumper.represent_str(str(data))



# stdlib
import os
import tempfile
from typing import Iterator

# 3rd party
import pytest
import requests
from domdf_python_tools.paths import PathPlus

# this package
from shippinglabel.checksum import check_sha256_hash, get_md5_hash, get_record_entry, get_sha256_hash


@pytest.fixture(scope="session")
def reference_file_a() -> Iterator[PathPlus]:
	with tempfile.TemporaryDirectory() as tmpdir:
		commit_sha = "4f632ed497bffa0cb50d714477de0cf731d34dc6"
		filename = "shippinglabel.svg"
		url = f"https://raw.githubusercontent.com/domdfcoding/shippinglabel/{commit_sha}/{filename}"

		tmpfile = PathPlus(tmpdir) / filename
		tmpfile.write_bytes(requests.get(url).content)

		yield tmpfile


def test_get_sha256_hash(reference_file_a: PathPlus):
	hash_ = get_sha256_hash(reference_file_a)
	assert hash_.hexdigest() == "83065efdedd381da9439b85a270ea9629f1ba46d9c7d7b1858bb70e54d5f664c"
	digest = b"\x83\x06^\xfd\xed\xd3\x81\xda\x949\xb8Z'\x0e\xa9b\x9f\x1b\xa4m\x9c}{\x18X\xbbp\xe5M_fL"
	assert hash_.digest() == digest


def test_get_sha256_hash_fp(reference_file_a: PathPlus):
	with open(reference_file_a, "rb") as fp:
		hash_ = get_sha256_hash(fp)

	assert hash_.hexdigest() == "83065efdedd381da9439b85a270ea9629f1ba46d9c7d7b1858bb70e54d5f664c"
	digest = b"\x83\x06^\xfd\xed\xd3\x81\xda\x949\xb8Z'\x0e\xa9b\x9f\x1b\xa4m\x9c}{\x18X\xbbp\xe5M_fL"
	assert hash_.digest() == digest


def test_get_md5_hash(reference_file_a: PathPlus):
	hash_ = get_md5_hash(reference_file_a)
	assert hash_.hexdigest() == "d7aadf6f1f20826aa227fa1bde281a41"
	assert hash_.digest() == b"\xd7\xaa\xdfo\x1f \x82j\xa2'\xfa\x1b\xde(\x1aA"


def test_get_md5_hash_fp(reference_file_a: PathPlus):
	with open(reference_file_a, "rb") as fp:
		hash_ = get_md5_hash(fp)

	assert hash_.hexdigest() == "d7aadf6f1f20826aa227fa1bde281a41"
	assert hash_.digest() == b"\xd7\xaa\xdfo\x1f \x82j\xa2'\xfa\x1b\xde(\x1aA"


def test_check_sha256_hash(reference_file_a: PathPlus):
	assert check_sha256_hash(
			reference_file_a,
			"83065efdedd381da9439b85a270ea9629f1ba46d9c7d7b1858bb70e54d5f664c",
			)
	assert check_sha256_hash(reference_file_a, get_sha256_hash(reference_file_a))


def test_check_sha256_hash_fp(reference_file_a: PathPlus):
	with open(reference_file_a, "rb") as fp:
		assert check_sha256_hash(
				fp,
				"83065efdedd381da9439b85a270ea9629f1ba46d9c7d7b1858bb70e54d5f664c",
				)

	with open(reference_file_a, "rb") as fp:
		assert check_sha256_hash(fp, get_sha256_hash(reference_file_a))


def test_get_record_entry(reference_file_a: PathPlus):
	entry = get_record_entry(reference_file_a, relative_to=reference_file_a.parent)
	assert entry == f"{os.path.basename(reference_file_a)},sha256=gwZe_e3TgdqUObhaJw6pYp8bpG2cfXsYWLtw5U1fZkw,154911"



# stdlib
import sys

# 3rd party
import pytest
from coincidence import only_windows
from coincidence.regressions import AdvancedDataRegressionFixture

# this package
from shippinglabel.requirements import list_requirements
from tests.test_requirements import min_311, only_36, only_37


@only_windows("Output differs on Windows")
@pytest.mark.parametrize(
		"py_version",
		[
				only_36,
				only_37,
				pytest.param(
						"3.8+",
						marks=pytest.mark.skipif(
								not ((3, 8) <= sys.version_info[:2] < (3, 11)),
								reason="Output differs on Python 3.8, 3.9, 3.10"
								)
						),
				min_311,
				]
		)
@pytest.mark.parametrize("depth", [-1, 0, 1, 2, 3])
def test_list_requirements_pytest(
		advanced_data_regression: AdvancedDataRegressionFixture,
		depth: int,
		py_version: str,
		):
	advanced_data_regression.check(list(list_requirements("pytest", depth=depth)))



# 3rd party
import pytest
from coincidence.regressions import AdvancedDataRegressionFixture
from packaging.utils import InvalidSdistFilename

# this package
from shippinglabel.sdist import NotAnSdistError, parse_sdist_filename


@pytest.mark.parametrize(
		"filename",
		[
				"wxPython-4.1.0.tar.gz",
				"PyGObject-3.38.0.tar.gz",
				"psutil-5.8.0.tar.gz",
				"tinycss-0.4.tar.gz",
				"cffi-1.14.4.tar.gz",
				"termcolor-1.1.0.tar.gz",
				"autodocsumm-0.2.2.tar.gz",
				"MarkupSafe-1.1.1.tar.gz",
				"astropy-4.2.tar.gz",
				"pyrsistent-0.17.3.tar.gz",
				"memoized-property-1.0.3.tar.gz",
				"PyNaCl-1.4.0.tar.gz",
				"pycairo-1.20.0.tar.gz",
				"strict-rfc3339-0.7.tar.gz",
				"autoflake-1.4.tar.gz",
				"Pillow-8.1.0.tar.gz",
				"cairocffi-1.2.0.tar.gz",
				"quantities-0.12.4.tar.gz",
				"pyerfa-1.7.1.1.tar.gz",
				"matplotlib-3.3.3.tar.gz",
				"fastimport-0.9.8.tar.gz",
				"orderedset-2.0.3.tar.gz",
				"PyYAML-5.3.1.tar.gz",
				"pandas-1.2.0.tar.gz",
				"Cython-0.29.14.tar.gz",
				"PyGObject-stubs-0.0.2.tar.gz",
				"dulwich-0.20.15.tar.gz",
				"tox-pip-version-0.0.7.tar.gz",
				"numpy-1.16.5.zip",
				]
		)
def test_parse_sdist_filename(filename: str, advanced_data_regression: AdvancedDataRegressionFixture):
	parsed_filename = parse_sdist_filename(filename)
	advanced_data_regression.check(parsed_filename)
	assert str(parsed_filename) == filename


def test_parse_sdist_filename_wheel():
	# Test passing in a wheel
	with pytest.raises(NotAnSdistError, match="'autopep8-1.5.4-py2.py3-none-any.whl' is a wheel."):
		parse_sdist_filename("autopep8-1.5.4-py2.py3-none-any.whl")


def test_parse_sdist_filename_invalid():
	with pytest.raises(InvalidSdistFilename, match="Invalid sdist filename: 'autopep8-.tar.bz2'"):
		parse_sdist_filename("autopep8-.tar.bz2")



# 3rd party
import pytest
from coincidence.regressions import AdvancedDataRegressionFixture
from domdf_python_tools.paths import PathPlus

# this package
from shippinglabel import no_dev_versions, no_pre_versions, normalize, normalize_keep_dot, read_pyvenv


@pytest.mark.parametrize(
		"name, expected",
		[
				("foo", "foo"),
				("bar", "bar"),
				("baz", "baz"),
				("baz-extensions", "baz-extensions"),
				("baz_extensions", "baz-extensions"),
				("baz.extensions", "baz-extensions"),
				]
		)
def test_normalize(name: str, expected: str):
	assert normalize(name) == expected


@pytest.mark.parametrize(
		"name, expected",
		[
				("foo", "foo"),
				("bar", "bar"),
				("baz", "baz"),
				("baz-extensions", "baz-extensions"),
				("baz_extensions", "baz-extensions"),
				("baz.extensions", "baz.extensions"),
				]
		)
def test_normalize_keep_dot(name: str, expected: str):
	assert normalize_keep_dot(name) == expected


def test_no_dev_versions():
	assert no_dev_versions(["3.6", "3.7", "3.8"]) == ["3.6", "3.7", "3.8"]
	assert no_dev_versions(["3.6", "3.7", "3.8", "3.9-dev"]) == ["3.6", "3.7", "3.8"]
	assert no_dev_versions(["3.6", "3.7", "3.8", "pypy3"]) == ["3.6", "3.7", "3.8", "pypy3"]


def test_no_pre_versions():
	assert no_pre_versions(["3.6", "3.7", "3.8"]) == ["3.6", "3.7", "3.8"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "3.9-dev"]) == ["3.6", "3.7", "3.8"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "pypy3"]) == ["3.6", "3.7", "3.8", "pypy3"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "3.10.0alpha1"]) == ["3.6", "3.7", "3.8"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "3.10.0beta3"]) == ["3.6", "3.7", "3.8"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "3.10.0rc2"]) == ["3.6", "3.7", "3.8"]
	assert no_pre_versions(["3.6", "3.7", "3.8", "3.10.0post1"]) == ["3.6", "3.7", "3.8", "3.10.0post1"]


def test_read_pyvenv(tmp_pathplus: PathPlus, advanced_data_regression: AdvancedDataRegressionFixture):
	(tmp_pathplus / "pyvenv.cfg").write_text(
			'\n'.join([
					"home = /usr",
					"implementation = CPython",
					"version_info = 3.8.5.final.0",
					"virtualenv = 20.2.2",
					"include-system-site-packages = false",
					"base-prefix = /usr",
					"base-exec-prefix = /usr",
					"base-executable = /usr/bin/python3",
					"prompt = (shippinglabel) ",
					"repo_helper_devenv = 0.3.0",
					])
			)

	venv_config = read_pyvenv(tmp_pathplus)
	assert venv_config["prompt"] == "(shippinglabel) "

	advanced_data_regression.check(venv_config)



# stdlib
import sys
from typing import List, Sequence, Union

# 3rd party
import pytest
from coincidence.regressions import AdvancedDataRegressionFixture
from coincidence.selectors import min_version, not_windows, only_version
from domdf_python_tools.paths import PathPlus
from packaging.requirements import Requirement
from packaging.specifiers import Specifier, SpecifierSet
from typing_extensions import Literal

# this package
from shippinglabel.requirements import (
		ComparableRequirement,
		combine_requirements,
		list_requirements,
		parse_pyproject_dependencies,
		parse_pyproject_extras,
		parse_requirements,
		read_requirements,
		resolve_specifiers
		)


class TestComparableRequirement:

	@pytest.fixture(scope="class")
	def req(self) -> ComparableRequirement:
		return ComparableRequirement('pytest==6.0.0; python_version <= "3.9"')

	@pytest.mark.parametrize(
			"other",
			[
					ComparableRequirement('pytest==6.0.0; python_version <= "3.9"'),
					ComparableRequirement("pytest==6.0.0"),
					ComparableRequirement("pytest"),
					ComparableRequirement("pytest[extra]"),
					Requirement('pytest==6.0.0; python_version <= "3.9"'),
					Requirement("pytest==6.0.0"),
					Requirement("pytest"),
					Requirement("pytest[extra]"),
					"pytest",
					]
			)
	def test_eq(self, req: ComparableRequirement, other: Union[str, ComparableRequirement]):
		assert req == req
		assert req == other

	@pytest.mark.parametrize(
			"other",
			[
					"pytest-rerunfailures",
					ComparableRequirement("pytest-rerunfailures"),
					ComparableRequirement("pytest-rerunfailures==1.2.3"),
					Requirement("pytest-rerunfailures"),
					Requirement("pytest-rerunfailures==1.2.3"),
					ComparableRequirement("pytest"),
					ComparableRequirement("pytest[extra]"),
					Requirement("pytest"),
					Requirement("pytest[extra]"),
					]
			)
	def test_gt(
			self,
			req: ComparableRequirement,
			other: Union[str, ComparableRequirement],
			):
		assert req < other

	@pytest.mark.parametrize(
			"other",
			[
					"apeye",
					ComparableRequirement("apeye"),
					ComparableRequirement("apeye==1.2.3"),
					Requirement("apeye"),
					Requirement("apeye==1.2.3"),
					]
			)
	def test_lt(
			self,
			req: ComparableRequirement,
			other: Union[str, ComparableRequirement],
			):
		assert req > other

	@pytest.mark.parametrize(
			"other",
			[
					"pytest-rerunfailures",
					ComparableRequirement("pytest-rerunfailures"),
					ComparableRequirement("pytest-rerunfailures==1.2.3"),
					ComparableRequirement('pytest==6.0.0; python_version <= "3.9"'),
					Requirement("pytest-rerunfailures"),
					Requirement("pytest-rerunfailures==1.2.3"),
					Requirement('pytest==6.0.0; python_version <= "3.9"'),
					ComparableRequirement("pytest==6.0.0"),
					ComparableRequirement("pytest"),
					ComparableRequirement("pytest[extra]"),
					Requirement("pytest==6.0.0"),
					Requirement("pytest"),
					Requirement("pytest[extra]"),
					"pytest",
					]
			)
	def test_ge(
			self,
			req: ComparableRequirement,
			other: Union[str, ComparableRequirement],
			):
		assert req <= other
		assert req <= req

	@pytest.mark.parametrize(
			"other",
			[
					"apeye",
					ComparableRequirement("apeye"),
					ComparableRequirement("apeye==1.2.3"),
					Requirement("apeye"),
					Requirement("apeye==1.2.3"),
					ComparableRequirement('pytest==6.0.0; python_version <= "3.9"'),
					ComparableRequirement("pytest==6.0.0"),
					ComparableRequirement("pytest"),
					ComparableRequirement("pytest[extra]"),
					Requirement('pytest==6.0.0; python_version <= "3.9"'),
					Requirement("pytest==6.0.0"),
					Requirement("pytest"),
					Requirement("pytest[extra]"),
					"pytest",
					]
			)
	def test_le(
			self,
			req: ComparableRequirement,
			other: Union[str, ComparableRequirement],
			):
		assert req >= other
		assert req >= req


def test_combine_requirements():
	reqs = [
			ComparableRequirement("foo"),
			ComparableRequirement("foo>2"),
			ComparableRequirement("foo>2.5"),
			ComparableRequirement("foo==3.2.1"),
			ComparableRequirement("foo==3.2.3"),
			ComparableRequirement("foo==3.2.5"),
			]

	assert combine_requirements(reqs) == [Requirement("foo==3.2.1,==3.2.3,==3.2.5,>2.5")]
	assert str(combine_requirements(reqs)[0]) == "foo==3.2.1,==3.2.3,==3.2.5,>2.5"
	assert str(combine_requirements(reqs)[0].specifier) == "==3.2.1,==3.2.3,==3.2.5,>2.5"


def test_combine_requirements_no_mutation():
	# Input objects should not be mutated
	reqs = [
			ComparableRequirement("foo>2"),
			ComparableRequirement("foo>2.5"),
			]
	assert combine_requirements(reqs) == [ComparableRequirement("foo>2.5")]

	assert reqs == [
			ComparableRequirement("foo>2"),
			ComparableRequirement("foo>2.5"),
			]


def test_combine_requirements_duplicates():
	reqs = [
			ComparableRequirement('typing-extensions>=3.6.4; python_version < "3.8"'),
			ComparableRequirement("typing-extensions>=3.7.4.3"),
			ComparableRequirement("typing-extensions>=3.7.4.3"),
			ComparableRequirement("typing-extensions>=3.7.4.3"),
			ComparableRequirement("typing-extensions>=3.7.4.3"),
			ComparableRequirement("typing-extensions>=3.7.4.1"),
			ComparableRequirement("typing-extensions>=3.7.4"),
			ComparableRequirement('typing-extensions; python_version < "3.8"'),
			]

	combined_reqs = combine_requirements(reqs)
	assert len(combined_reqs) == 2
	assert combined_reqs[1] == ComparableRequirement("typing-extensions>=3.7.4.3")
	assert combined_reqs[0] == ComparableRequirement('typing-extensions>=3.6.4; python_version < "3.8"')

	reqs.append(reqs.pop(0))

	combined_reqs = combine_requirements(reqs)
	assert len(combined_reqs) == 2
	assert combined_reqs[0] == ComparableRequirement("typing-extensions>=3.7.4.3")
	assert combined_reqs[1] == ComparableRequirement('typing-extensions>=3.6.4; python_version < "3.8"')


def test_combine_requirements_differing_precision():
	reqs = [
			ComparableRequirement("lockfile>=0.9"),
			ComparableRequirement("lockfile>=0.9"),
			ComparableRequirement("lockfile>=0.12.2"),
			]

	assert combine_requirements(reqs) == [Requirement("lockfile>=0.12.2")]


@pytest.mark.parametrize(
		"reqs, combined",
		[
				(
						[
								ComparableRequirement('numpy==1.19.3; platform_system == "Windows"'),
								ComparableRequirement('numpy>=1.19.1; platform_system != "Windows"')
								],
						[
								ComparableRequirement('numpy==1.19.3; platform_system == "Windows"'),
								ComparableRequirement('numpy>=1.19.1; platform_system != "Windows"')
								],
						),
				(
						[
								ComparableRequirement('numpy==1.19.3; platform_system == "Windows"'),
								ComparableRequirement("numpy>=1.19.1"),
								],
						[
								ComparableRequirement('numpy==1.19.3; platform_system == "Windows"'),
								ComparableRequirement("numpy>=1.19.1"),
								],
						),
				(
						[ComparableRequirement("numpy==1.19.3"), ComparableRequirement("numpy>=1.19.1")],
						[ComparableRequirement("numpy==1.19.3")],
						),
				(
						[ComparableRequirement("numpy<=1.19.3"), ComparableRequirement("numpy==1.19.1")],
						[ComparableRequirement("numpy==1.19.1")],
						),
				(
						[ComparableRequirement("numpy<=1.19.3"), ComparableRequirement("numpy<1.19.1")],
						[ComparableRequirement("numpy<1.19.1")],
						),
				(
						[ComparableRequirement("numpy>1.2.3"), ComparableRequirement("numpy>=1.2.2")],
						[ComparableRequirement("numpy>1.2.3")],
						),
				]
		)
def test_combine_requirements_markers(
		reqs: List[ComparableRequirement],
		combined: List[ComparableRequirement],
		):
	assert combine_requirements(reqs) == combined


@pytest.mark.parametrize(
		"specifiers, resolved",
		[
				([Specifier(">1.2.3"), Specifier(">=1.2.2"), Specifier("<2")], SpecifierSet(">1.2.3,<2")),
				([Specifier(">1.2.3"), Specifier(">=1.2.2")], SpecifierSet(">1.2.3")),
				([Specifier(">=1.2.2"), Specifier("<2")], SpecifierSet(">=1.2.2,<2")),
				([Specifier(">1.2.3"), Specifier("<2")], SpecifierSet(">1.2.3,<2")),
				([Specifier("<1.2.2"), Specifier("<=1.2.3"), Specifier(">2")], SpecifierSet("<1.2.2,>2")),
				([Specifier("<1.2.2"), Specifier("<=1.2.3")], SpecifierSet("<1.2.2")),
				([Specifier("<=1.2.3"), Specifier(">2")], SpecifierSet("<=1.2.3,>2")),
				([Specifier("<1.2.2"), Specifier(">2")], SpecifierSet("<1.2.2,>2")),
				]
		)
def test_resolve_specifiers(specifiers: List[Specifier], resolved: SpecifierSet):
	assert resolve_specifiers(specifiers) == resolved


requirements_a = [
		"autodocsumm>=0.2.0",
		"default-values>=0.2.0",
		"domdf-sphinx-theme>=0.1.0",
		"extras-require>=0.2.0",
		"repo-helper-sphinx-theme>=0.0.2",
		"seed-intersphinx-mapping>=0.1.1",
		"sphinx>=3.0.3",
		"ruamel-yaml>=0.16.12",
		"sphinx-click>=2.5.0",
		"sphinx-copybutton>=0.2.12",
		"sphinx-notfound-page>=0.5",
		"sphinx-prompt>=1.1.0",
		"sphinx-tabs>=1.1.13",
		"sphinx-toolbox>=1.7.1",
		"sphinxcontrib-autoprogram>=0.1.5",
		"sphinxcontrib-httpdomain>=1.7.0",
		"sphinxemoji>=0.1.6",
		"toctree-plus>=0.0.4",
		]

requirements_b = [
		"autodocsumm>=0.2.0",
		"default-values>=0.2.0",
		"domdf-sphinx-theme>=0.1.0",
		"domdf-sphinx-theme>=0.1.0",
		"extras-require>=0.2.0",
		"repo-helper-sphinx-theme>=0.0.2",
		"seed-intersphinx-mapping>=0.1.1",
		"sphinx>=3.0.3",
		"sphinx-click>=2.5.0",
		"sphinx-copybutton>=0.2.12",
		"sphinx-copybutton>=0.2.12",
		"sphinx-notfound-page>=0.5",
		"sphinx-prompt>=1.1.0",
		"sphinx-tabs>=1.1.13",
		"sphinx-toolbox>=1.7.1",
		"ruamel.yaml>=0.16.12",
		"sphinxcontrib-autoprogram>=0.1.5",
		"sphinxcontrib-autoprogram>=0.1.5",
		"sphinxcontrib-httpdomain>=1.7.0",
		"sphinxemoji>=0.1.6",
		"toctree-plus>=0.0.4",
		"toctree-plus>=0.0.3",
		]

requirements_c = [
		'numpy==1.19.3; platform_system == "Windows"',
		'numpy>=1.19.1; platform_system != "Windows"',
		]


@pytest.mark.parametrize(
		"requirements",
		[
				pytest.param(requirements_a, id='a'),
				pytest.param(requirements_b, id='b'),
				pytest.param(requirements_c, id='c'),
				]
		)
def test_read_requirements(
		tmp_pathplus: PathPlus,
		advanced_data_regression: AdvancedDataRegressionFixture,
		requirements: List[str],
		):
	(tmp_pathplus / "requirements.txt").write_lines(requirements)
	advanced_data_regression.check([
			str(x) for x in sorted(read_requirements(tmp_pathplus / "requirements.txt")[0])
			])


@pytest.mark.parametrize(
		"requirements",
		[
				pytest.param(requirements_a, id='a'),
				pytest.param(requirements_b, id='b'),
				pytest.param(requirements_c, id='c'),
				pytest.param(iter(requirements_a), id="iter(a)"),
				pytest.param(iter(requirements_b), id="iter(b)"),
				pytest.param(iter(requirements_c), id="iter(c)"),
				pytest.param(set(requirements_a), id="set(a)"),
				pytest.param(set(requirements_b), id="set(b)"),
				pytest.param(set(requirements_c), id="set(c)"),
				pytest.param(tuple(requirements_a), id="tuple(a)"),
				pytest.param(tuple(requirements_b), id="tuple(b)"),
				pytest.param(tuple(requirements_c), id="tuple(c)"),
				]
		)
def test_parse_requirements(
		advanced_data_regression: AdvancedDataRegressionFixture,
		requirements: List[str],
		):
	advanced_data_regression.check([str(x) for x in sorted(parse_requirements(requirements)[0])])


@min_version("3.7", reason="Latest packaging is 3.7+")
def test_read_requirements_invalid(
		tmp_pathplus: PathPlus,
		advanced_data_regression: AdvancedDataRegressionFixture,
		):
	(tmp_pathplus / "requirements.txt").write_lines([
			"# another comment",
			"autodocsumm>=apples",
			"default-value---0.2.0",
			"domdf-sphinx-theme!!!0.1.0",
			"0.2.0",
			'',
			'',
			"https://bbc.co.uk",
			"toctree-plus>=0.0.4",
			"   # a comment",
			])

	with pytest.warns(UserWarning) as record:
		requirements, comments = read_requirements(tmp_pathplus / "requirements.txt")

	# check that only one warning was raised
	assert len(record) == 3

	# check that the messages match
	for idx, warning in enumerate([
		"Ignored invalid requirement 'autodocsumm>=apples'",
		"Ignored invalid requirement 'domdf-sphinx-theme!!!0.1.0'",
		"Ignored invalid requirement 'https://bbc.co.uk'",
	]):
		assert record[idx].message.args[0] == warning  # type: ignore[index]  # Issue with "Self"

	advanced_data_regression.check([str(x) for x in sorted(requirements)])
	assert comments == ["# another comment", "   # a comment"]


@only_version("3.6", reason="Latest packaging is 3.7+")
def test_read_requirements_invalid_py36(
		tmp_pathplus: PathPlus,
		advanced_data_regression: AdvancedDataRegressionFixture,
		):
	(tmp_pathplus / "requirements.txt").write_lines([
			"# another comment",
			"autodocsumm>=apples",
			"default-value---0.2.0",
			"domdf-sphinx-theme!!!0.1.0",
			"0.2.0",
			'',
			'',
			"https://bbc.co.uk",
			"toctree-plus>=0.0.4",
			"   # a comment",
			])

	with pytest.warns(UserWarning) as record:
		requirements, comments = read_requirements(tmp_pathplus / "requirements.txt")

	# check that only one warning was raised
	assert len(record) == 3

	# check that the messages match
	for idx, warning in enumerate([
		"Creating a LegacyVersion has been deprecated and will be removed in the next major release",
		"Ignored invalid requirement 'domdf-sphinx-theme!!!0.1.0'",
		"Ignored invalid requirement 'https://bbc.co.uk'",
	]):
		assert record[idx].message.args[0] == warning  # type: ignore[index]  # Issue with "Self"

	advanced_data_regression.check([str(x) for x in sorted(requirements)])
	assert comments == ["# another comment", "   # a comment"]


def test_sort_mixed_requirements():

	requirements: Sequence[Union[str, ComparableRequirement]] = [
			"urllib3",
			ComparableRequirement("six==1.15.0"),
			"botocore",
			ComparableRequirement("requests>=2.19.1"),
			"python-dateutil",
			]

	assert sorted(requirements) == [
			"botocore",
			"python-dateutil",
			ComparableRequirement("requests>=2.19.1"),
			ComparableRequirement("six==1.15.0"),
			"urllib3",
			]


def test_comparable_requirement():
	assert ComparableRequirement("foo") != ComparableRequirement("bar")
	assert ComparableRequirement("foo") == ComparableRequirement("foo")
	assert ComparableRequirement("foo>=1.2.3") == ComparableRequirement("foo >= 1.2.3")

	def req_with_marker() -> ComparableRequirement:
		return ComparableRequirement('importlib-metadata>=1.5.0; python_version < "3.8"')

	def req_without_marker() -> ComparableRequirement:
		return ComparableRequirement("importlib-metadata>=1.5.0")

	def req_with_different_marker() -> ComparableRequirement:
		return ComparableRequirement('importlib-metadata>=1.5.0; python_version < "3.10"')

	assert req_with_marker() == req_with_marker()
	assert req_with_marker() is not req_with_marker()
	assert req_without_marker() is not req_without_marker()
	assert req_with_marker() != req_with_different_marker()

	assert "importlib-metadata" in [req_with_marker()]
	assert req_without_marker() in [req_with_marker()]
	assert req_with_marker() in [req_with_marker()]

	assert "importlib-metadata" in (req_with_marker(), )
	assert req_without_marker() in (req_with_marker(), )
	assert req_with_marker() in (req_with_marker(), )

	assert {req_without_marker(), req_without_marker()} == {req_without_marker()}
	assert {req_with_marker(), req_with_marker()} == {req_with_marker()}

	assert hash(req_with_marker()) == hash(req_with_marker())
	assert hash(req_with_marker()) != hash(req_without_marker())

	assert req_without_marker() not in {req_with_marker()}
	assert req_with_marker() in {req_with_marker()}

	assert req_without_marker() != "123foo?"


only_36 = pytest.param("3.6", marks=only_version((3, 6), reason="Output differs on Python 3.6"))
only_37 = pytest.param("3.7", marks=only_version((3, 7), reason="Output differs on Python 3.7"))
only_38 = pytest.param("3.8", marks=only_version((3, 8), reason="Output differs on Python 3.8"))
min_38 = pytest.param("3.8+", marks=min_version((3, 8), reason="Output differs on Python 3.8+"))
min_311 = pytest.param("3.11+", marks=min_version((3, 11), reason="Output differs on Python 3.11+"))
only_39 = pytest.param("3.9", marks=only_version((3, 9), reason="Output differs on Python 3.9"))
only_310 = pytest.param("3.10", marks=only_version((3, 10), reason="Output differs on Python 3.10"))


@not_windows("Output differs on Windows")
@pytest.mark.parametrize("py_version", [
		only_36,
		only_37,
		only_38,
		only_39,
		only_310,
		])
@pytest.mark.parametrize(
		"library", [
				"shippinglabel",
				"apeye",
				"cachecontrol[filecache]",
				"domdf-python-tools",
				"domdf_python_tools",
				]
		)
@pytest.mark.parametrize("depth", [-1, 0, 1, 2, 3])
# @pytest.mark.parametrize("depth", [3])
def test_list_requirements(
		advanced_data_regression: AdvancedDataRegressionFixture,
		library: str,
		depth: int,
		py_version: str,
		):
	advanced_data_regression.check(list(list_requirements(library, depth=depth)))


@not_windows("Output differs on Windows")
@pytest.mark.parametrize(
		"py_version",
		[
				only_36,
				only_37,
				pytest.param(
						"3.8+",
						marks=pytest.mark.skipif(
								not ((3, 8) <= sys.version_info[:2] < (3, 11)),
								reason="Output differs on Python 3.8, 3.9, 3.10"
								)
						),
				min_311,
				]
		)
@pytest.mark.parametrize("depth", [-1, 0, 1, 2, 3])
# @pytest.mark.parametrize("depth", [3])
def test_list_requirements_pytest(
		advanced_data_regression: AdvancedDataRegressionFixture,
		depth: int,
		py_version: str,
		):
	advanced_data_regression.check(list(list_requirements("pytest", depth=depth)))


@pytest.fixture()
def pyproject_toml(tmp_pathplus: PathPlus) -> PathPlus:

	filename = (tmp_pathplus / "pyproject.toml")
	filename.write_lines([
			"[build-system]",
			'requires = [ "setuptools>=40.6.0", "wheel>=0.34.2",]',
			'build-backend = "setuptools.build_meta"',
			'',
			"[project]",
			"dependencies = [",
			'  "httpx",',
			'  "gidgethub[httpx]>4.0.0",',
			"  \"django>2.1; os_name != 'nt'\",",
			"  \"django>2.0; os_name == 'nt'\"",
			']',
			'',
			"[project.optional-dependencies]",
			"test = [",
			'  "pytest < 5.0.0",',
			'  "pytest-cov[all]"',
			']',
			"[tool.flit.metadata]",
			"requires = [",
			'\t"requests >=2.6",',
			"\t\"configparser; python_version == '2.7'\",",
			']',
			'',
			"[tool.flit.metadata.requires-extra]",
			"test = [",
			'\t"pytest >=2.7.3",',
			'\t"pytest-cov",',
			']',
			])

	return filename


@pytest.mark.parametrize("flavour", ["auto", "pep621", "flit"])
def test_parse_pyproject_dependencies(
		pyproject_toml: PathPlus,
		advanced_data_regression: AdvancedDataRegressionFixture,
		flavour: Literal["auto", "pep621", "flit"],
		):
	deps = parse_pyproject_dependencies(pyproject_toml, flavour)
	advanced_data_regression.check(sorted(str(x) for x in deps))


@pytest.mark.parametrize("flavour", ["auto", "pep621", "flit"])
def test_parse_pyproject_extras(
		pyproject_toml: PathPlus,
		advanced_data_regression: AdvancedDataRegressionFixture,
		flavour: Literal["auto", "pep621", "flit"],
		):
	extras = parse_pyproject_extras(pyproject_toml, flavour)
	advanced_data_regression.check({k: sorted(str(x) for x in v) for k, v in extras.items()})
